import operator
import numpy as np
from staircase.core.tools import _sanitize_binary_operands
from staircase.util._decorators import Appender
from staircase.core.ops import docstrings
import staircase as sc


@Appender(docstrings.negate_docstring, join="\n", indents=1)
def negate(self):
    new_instance = self.copy()
    for key, delta in new_instance._items():
        new_instance[key] = -delta
    new_instance.cached_cumulative = None
    return new_instance


def _make_add_subtract_func(docstring, op):
    @Appender(docstring, join="\n", indents=1)
    def func(self, other):
        new_instance, other = _sanitize_binary_operands(self, other)
        for key, value in other._items():
            new_instance[key] = op(self._get(key, 0), value)
        new_instance._reduce()
        new_instance.use_dates = self.use_dates or other.use_dates
        new_instance.cached_cumulative = None
        return new_instance

    return func


add = _make_add_subtract_func(docstrings.add_docstring, operator.add)
subtract = _make_add_subtract_func(docstrings.subtract_docstring, operator.sub)


def _make_mul_div_func(docstring, op):

    is_divide = op == np.divide

    @Appender(docstring, join="\n", indents=1)
    def func(self, other):
        self_copy, other_copy = _sanitize_binary_operands(self, other, copy_other=True)
        if is_divide and not bool(other.make_boolean()):
            raise ZeroDivisionError(
                "Divisor Stairs instance must not be zero-valued at any point"
            )

        self_keys = self._keys()
        other_keys = other._keys()
        self_copy._layer_multiple(other_keys, None, [0] * len(other_keys))
        other_copy._layer_multiple(self_keys, None, [0] * len(self_keys))

        multiplied_cumulative_values = op(
            self_copy._cumulative().values(), other_copy._cumulative().values()
        )
        new_instance = sc.Stairs.from_cumulative(
            dict(zip(self_copy._keys(), multiplied_cumulative_values)),
            use_dates=self.use_dates,
            tz=self.tz,
        )
        new_instance._reduce()
        return new_instance

    return func


multiply = _make_mul_div_func(docstrings.multiply_docstring, np.multiply)
divide = _make_mul_div_func(docstrings.divide_docstring, np.divide)



from pandas import (
    Series,
    Timedelta,
    TimedeltaIndex,
    Timestamp,
    to_datetime,
    to_timedelta,
)
import numpy as np

origin = to_datetime("2000-1-1")


def check_binop_timezones(first, second):
    assert first.tz == second.tz, "operands must have the same timezone, or none at all"


def _convert_date_to_float(val, tz=None):
    if val is None:
        return None
    if hasattr(val, "__iter__"):
        if not isinstance(val, Series):
            val = Series(val)
        deltas = TimedeltaIndex(val - origin.tz_localize(tz))
        return list(deltas / Timedelta(1, "h"))
    return (val - origin.tz_localize(tz)) / Timedelta(1, "h")


def _convert_float_to_date(val, tz=None):
    if hasattr(val, "__iter__"):
        if not isinstance(val, np.ndarray):
            val = np.array(val)
        return list(to_timedelta(val * 3600, unit="s") + origin.tz_localize(tz))
    return to_timedelta(val * 3600, unit="s") + origin.tz_localize(tz)


def _maybe_convert_from_timedeltas(iterable):
    def maybe_convert(x):
        return x / Timedelta("1 hr") if isinstance(x, Timedelta) else x

    return [maybe_convert(x) for x in iterable]


def _maybe_convert_from_timestamps(iterable, tz):
    def maybe_convert(x):
        return (
            _convert_date_to_float(x, tz)
            if isinstance(x, Timestamp) or hasattr(x, "__iter__")
            else x
        )

    return [maybe_convert(x) for x in iterable]


def _using_dates(collection):
    def dict_use_dates():
        s = next(iter(collection.values()))
        return s.use_dates, s.tz

    def series_use_dates():
        s = collection.values[0]
        return s.use_dates, s.tz

    def array_use_dates():
        s = collection[0]
        return s.use_dates, s.tz

    for func in (dict_use_dates, series_use_dates, array_use_dates):
        try:
            return func()
        except Exception:
            pass
    raise TypeError(
        "Could not determine if Stairs collection is using dates.  Collection should be a tuple, list, numpy array, dict or pandas.Series."
    )



from staircase.core.ops.arithmetic import add, subtract, multiply, negate, divide
from staircase.core.ops.relational import eq, ne, lt, gt, le, ge, identical
from staircase.core.ops.logical import invert, logical_and, logical_or, make_boolean


def add_operations(cls):

    cls.negate = negate

    cls.add = add
    cls.subtract = subtract
    cls.multiply = multiply
    cls.divide = divide
    cls.eq = eq
    cls.ne = ne
    cls.lt = lt
    cls.gt = gt
    cls.le = le
    cls.ge = ge

    cls.__neg__ = negate
    cls.__add__ = add
    cls.__sub__ = subtract
    cls.__mul__ = multiply
    cls.__truediv__ = divide
    cls.__eq__ = eq
    cls.__ne__ = ne
    cls.__lt__ = lt
    cls.__gt__ = gt
    cls.__le__ = le
    cls.__ge__ = ge

    cls.invert = invert
    cls.logical_and = logical_and
    cls.logical_or = logical_or
    cls.make_boolean = make_boolean

    cls.__or__ = logical_or
    cls.__and__ = logical_and
    cls.__invert__ = invert

    cls.identical = identical



class Defaults:
    def __init__(self):
        self.tz_default = None
        self.use_dates_default = False
        self.dict = {
            "tz": None,
            "use_dates": False,
        }

    def set_default(self, name, val):
        self.dict.update({name: val})

    def get_default(self, name):
        return self.dict[name]


default = Defaults()



import operator
import numpy as np
import staircase as sc


def _make_extremum_func(extrema):
    """
    Used to generate _min_pair and _max_pair respectively to cut code duplication

    Parameters
    ----------
    extrema : {"minimum", "maximum"}

    Returns:
        Callable
    """
    comparator = {"minimum": operator.gt, "maximum": operator.lt}[extrema]

    def extremum_func(stairs1, stairs2):
        f"""
        Calculates the {extrema} of two Stairs objects.  It can be thought of as calculating the {extrema} of two step functions.

        Parameters:
            stairs1 (Stairs)
            stairs2 (Stairs)

        Returns:
            Stairs: the result of the calculation
        """
        assert isinstance(stairs1, sc.Stairs) and isinstance(
            stairs2, sc.Stairs
        ), "Arguments must be both of type Stairs."
        assert stairs1.tz == stairs2.tz
        new_instance = stairs1 - stairs2
        cumulative = new_instance._cumulative()
        for key, value in cumulative.items():
            if comparator(value, 0):
                cumulative[key] = 0
        deltas = [cumulative.values()[0]]
        deltas.extend(np.subtract(cumulative.values()[1:], cumulative.values()[:-1]))
        new_instance = sc.Stairs(
            dict(zip(new_instance._keys(), deltas)),
            use_dates=stairs1.use_dates or stairs2.use_dates,
            tz=stairs1.tz,
        )
        return new_instance + stairs2

    return extremum_func


_min_pair = _make_extremum_func("minimum")
_max_pair = _make_extremum_func("maximum")



_std_var_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.{func}()
    {result1}
    >>> s1.{func}(lower=0, upper=6)
    {result2}
"""

var_example = _std_var_example.format(
    func="var", result1="0.6875", result2="0.4722222222222224",
)

std_example = _std_var_example.format(
    func="std", result1="0.82915619758885", result2="0.6871842709362769",
)

integrate_mean_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.{func}(3, 4.5)
    {result}
"""

integral_and_mean_example = integrate_mean_example.format(
    func="get_integral_and_mean", result="(0.5, 0.3333333333333333)",
)

integrate_example = integrate_mean_example.format(func="integrate", result="0.5",)

mean_example = integrate_mean_example.format(func="mean", result="0.3333333333333333",)

mode_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s2.plot()
    >>> s2.mode()
    -1.0
    >>> s2.mode(0,3)
    0.5
"""


median_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.median()
    0.5
    >>> s1.median(2,5)
    0.0
"""

percentile_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.percentile(70)
    1.0
    >>> s1.percentile(25)
    -0.5
    >>> s1.percentile(40, lower=1.5, upper=3.5)
    0.0
"""

percentile_stairs_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5))
    >>> s2.plot(axes[0])
    >>> axes[0].set_title("s2")
    >>> s2_percentiles = s2.percentile_stairs()
    >>> s2_percentiles.plot(axes[1])
    >>> axes[0].set_title("s2 percentiles")
    >>> s2_percentiles(55)
    0.0
    >>> s2_percentiles(75)
    0.5
"""

hist_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()

.. plot::
    :context: close-figs

    >>> s1.hist()
    [-1, 0)    0.25
    [0, 1)     0.25
    [1, 2)     0.50
    dtype: float64

    >>> s1.hist(closed='right')
    (-2, -1]    0.25
    (-1, 0]     0.25
    (0, 1]      0.50
    dtype: float64

    >>> s1.hist(2, 4.5)
    [-1, 0)    0.2
    [0, 1)     0.4
    [1, 2)     0.4
    dtype: float64

    >>> s1.hist(bin_edges=(-1,1,3))
    [-1, 1)    0.5
    [1, 3)     0.5
    dtype: float64

    >>> s1.hist(bin_edges=(-1, 1))
    [-1, 1)    0.5
    dtype: float64
"""

ecdf_stairs_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> s2.plot()

.. plot::
    :context: close-figs

    >>> ecdf = s2.ecdf_stairs(1,5)
    >>> ecdf.plot()
    >>> plt.show()


    >>> print(f'{ecdf(0):.2%} of values, for s2 between 1 and 5, are less than or equal to 0')
    75.00% of values, for s2 between 1 and 5, are less than or equal to 0

    >>> print(f'{ecdf(0, how="left"):.2%} of values, for s2 between 1 and 5, are strictly less than 0')
    50.00% of values, for s2 between 1 and 5, are strictly less than 0

    >>> print(f'{ecdf(0.2) - ecdf(-1):.2%} of values, for s2 between 1 and 5, are in (-1, 0.2]')
    25.00% of values, for s2 between 1 and 5, are in (-1, 0.2]

    >>> print(f'{ecdf(0.2, how="left") - ecdf(-1, how="left"):.2%} of values, for s2 between 1 and 5, are in [-1, 0.2)')
    75.00% of values, for s2 between 1 and 5, are in [-1, 0.2)
"""

hist_from_ecdf_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> import staircase as sc
    >>> s1_ecdf_stairs = s1.ecdf_stairs()
    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(8,5), sharey=False, sharex=False)
    >>> for ax, title, stair_instance in zip(axes, ("s1", "s1 ecdf"), (s1, s1_ecdf_stairs)):
    ...     stair_instance.plot(ax)
    ...     ax.set_title(title)
    >>> plt.show()

    >>> sc.hist_from_ecdf(s1_ecdf_stairs)
    [-1, 0)    0.25
    [0, 1)     0.25
    [1, 2)     0.50
    dtype: float64

    >>> sc.hist_from_ecdf(s1_ecdf_stairs, bin_edges=(-1,1,3))
    [-1, 1)    0.5
    [1, 3)     0.5
    dtype: float64
"""


def _get_example(calculation):

    return {
        "mean": mean_example,
        "integrate": integrate_example,
        "integral_and_mean": integral_and_mean_example,
        "percentile": percentile_stairs_example,
        "median": median_example,
        "mode": mode_example,
        "var": var_example,
        "std": std_example,
    }[calculation]


_calc_map = {
    "integrate": "integral",
    "integral_and_mean": "integral, and the mean",
    "percentile": "x-th percentile",
    "var": "variance",
    "std": "standard deviation",
}

_see_also_map = {
    "mean": "Stairs.rolling_mean, Stairs.get_integral_and_mean, Stairs.median, Stairs.mode",
    "integrate": "Stairs.get_integral_and_mean",
    "integral_and_mean": "Stairs.integrate, Stairs.mean",
    "percentile": "Stairs.median, Stairs.percentile_stairs",
    "median": "Stairs.mean, Stairs.mode, Stairs.percentile, Stairs.percentile_stairs",
    "mode": "Stairs.mean, Stairs.median",
    "var": "Stairs.std",
    "std": "Stairs.var",
}

_common_docstring = """
Calculates the {{calc}} of the step function.

Parameters
----------
lower : int, float or pandas.Timestamp
    lower bound of the interval on which to perform the calculation
upper : int, float or pandas.Timestamp
    upper bound of the interval on which to perform the calculation

Returns
-------
{return_type}
     The {{calc}} of the step function

See Also
--------
{{see_also}}

{{example}}
"""

_float_return_docstring = _common_docstring.format(return_type="float")
_tuple_return_docstring = _common_docstring.format(return_type="tuple")


def _gen_docstring(calculation):

    calc = _calc_map.get(calculation, calculation)
    see_also = _see_also_map[calculation]
    example = _get_example(calculation)

    if calculation == "integral_and_mean":
        docstring = _tuple_return_docstring
    else:
        docstring = _float_return_docstring
    return docstring.format(calc=calc, see_also=see_also, example=example,)


mean_docstring = _gen_docstring("mean")
integrate_docstring = _gen_docstring("integrate")
percentile_docstring = _gen_docstring("percentile")
median_docstring = _gen_docstring("median")
mode_docstring = _gen_docstring("mode")
var_docstring = _gen_docstring("var")
std_docstring = _gen_docstring("std")
integral_and_mean_docstring = _gen_docstring("integral_and_mean")



import math
import warnings

import numpy as np
from pandas import IntervalIndex, Series

from staircase.util._decorators import Appender
from staircase.core.stats import docstrings
import staircase as sc

warnings.simplefilter("default")


@Appender(docstrings.percentile_stairs_example, join="\n", indents=1)
def percentile_stairs(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates a percentile function (and returns a corresponding Stairs instance)

    This method can be used for efficiency gains if substituting for multiple calls
    to percentile() with the same lower and upper parameters

    Parameters
    ----------
    lower : int, float or pandas.Timestamp, optional
        lower bound of the interval on which to perform the calculation
    upper : int, float or pandas.Timestamp, optional
        upper bound of the interval on which to perform the calculation

    Returns
    -------
    :class:`Stairs`
        An instance representing a percentile function

    See Also
    --------
    Stairs.percentile
    """
    temp_df = self.clip(lower, upper).to_dataframe()

    assert (
        temp_df.shape[0] >= 3
    ), "Step function composed only of infinite length intervals.  Provide bounds by 'lower' and 'upper' parameters"

    temp_df = (
        temp_df.iloc[1:-1]
        .assign(duration=lambda df: df.end - df.start)
        .groupby("value")
        .sum()
        .assign(duration=lambda df: np.cumsum(df.duration / df.duration.sum()))
        .assign(duration=lambda df: df.duration.shift())
        .fillna(0)
    )
    percentile_step_func = sc.Stairs()
    for start, end, value in zip(
        temp_df.duration.values,
        np.append(temp_df.duration.values[1:], 1),
        temp_df.index,
    ):
        percentile_step_func.layer(start * 100, end * 100, value)
    percentile_step_func[100] = 0
    return percentile_step_func


def percentile_Stairs(self, lower=float("-inf"), upper=float("inf")):
    """Deprecated.  Use Stairs.percentile_stairs."""
    warnings.warn(
        "Stairs.percentile_Stairs will be deprecated in version 2.0.0, use Stairs.percentile_stairs instead",
        PendingDeprecationWarning,
    )
    return percentile_stairs(self, lower, upper)


@Appender(docstrings.ecdf_stairs_example, join="\n", indents=1)
def ecdf_stairs(self, lower=float("-inf"), upper=float("inf")):
    """
    Calculates an `empirical cumulative distribution function <https://en.wikipedia.org/wiki/Empirical_distribution_function>`_
    for the corresponding step function values (and returns the result as a Stairs instance)

    Parameters
    ----------
    lower : int, float or pandas.Timestamp, optional
        lower bound of the step-function domain on which to perform the calculation
    upper : int, float or pandas.Timestamp, optional
        upper bound of the step-function domain to perform the calculation

    Returns
    -------
    :class:`Stairs`
        An instance representing an empirical cumulative distribution function for the step function values

    See Also
    --------
    staircase.hist_from_ecdf
    Stairs.hist
    """

    def _switch_first_key_to_zero(d):
        d[0] = d.get(0, 0) + d.pop(float("-inf"))
        return d

    _ecdf = _switch_first_key_to_zero(
        percentile_stairs(self, lower, upper)._sorted_dict.copy()
    )

    return sc.Stairs().layer(
        np.cumsum(list(_ecdf.values())[:-1]), None, np.diff(list(_ecdf.keys())) / 100
    )


@Appender(docstrings.hist_from_ecdf_example, join="\n", indents=1)
def hist_from_ecdf(ecdf, bin_edges=None, closed="left"):
    """
    Calculates a histogram from a Stairs instance corresponding to an
    `empirical cumulative distribution function <https://en.wikipedia.org/wiki/Empirical_distribution_function>`_.

    Such ecdf stair instances are returned from :meth:`Stairs.ecdf_stairs`.  This function predominantly exists
    to allow users to store the result of a ecdf stairs instance locally, and experiment with bin_edges without
    requiring the recalculation of the ecdf.

    Parameters
    ----------
    ecdf : :class:`Stairs`
        lower bound of the step-function domain on which to perform the calculation
    bin_edges : int, float, optional
        defines the bin edges for the histogram (it is the domain of the ecdf that is being binned).
        If not specified the bin_edges will be assumed to be the integers which cover the domain of the ecdf
    closed: {'left', 'right'}, default 'left'
        determines whether the bins, which are half-open intervals, are left-closed , or right-closed

    Returns
    -------
    :class:`pandas.Series`
        A Series, with a :class:`pandas.IntervalIndex`, representing the values of the histogram

    See Also
    --------
    Stairs.hist
    Stairs.ecdf_stairs
    """
    if bin_edges is None:
        round_func = math.floor if closed == "left" else math.ceil
        bin_edges = range(
            round_func(min(ecdf.step_changes().keys())) - (closed == "right"),
            round_func(max(ecdf.step_changes().keys())) + (closed == "left") + 1,
        )
    return Series(
        data=[
            ecdf(c2, how=closed) - ecdf(c1, how=closed)
            for c1, c2 in zip(bin_edges[:-1], bin_edges[1:])
        ],
        index=IntervalIndex.from_tuples(
            [(c1, c2) for c1, c2 in zip(bin_edges[:-1], bin_edges[1:])], closed=closed
        ),
        dtype="float64",
    )


@Appender(docstrings.hist_example, join="\n", indents=1)
def hist(self, lower=float("-inf"), upper=float("inf"), bin_edges=None, closed="left"):
    """
    Calculates a histogram for the corresponding step function values

    Parameters
    ----------
    lower : int, float or pandas.Timestamp, optional
        lower bound of the step-function domain on which to perform the calculation
    upper : int, float or pandas.Timestamp, optional
        upper bound of the step-function domain to perform the calculation
    bin_edges : array-like of int or float, optional
        defines the bin edges for the histogram (remember it is the step-function range that is being binned).
        If not specified the bin_edges will be assumed to be the integers which cover the step function range
    closed: {'left', 'right'}, default 'left'
        determines whether the bins, which are half-open intervals, are left-closed , or right-closed

    Returns
    -------
    :class:`pandas.Series`
        A Series, with a :class:`pandas.IntervalIndex`, representing the values of the histogram

    See Also
    --------
    staircase.hist_from_ecdf
    Stairs.ecdf_stairs
    """
    _ecdf = ecdf_stairs(self, lower, upper)
    return hist_from_ecdf(_ecdf, bin_edges, closed)



"""
Staircase
==============

Staircase is a MIT licensed library, written in pure-Python, for
modelling step functions. See :ref:`Getting Started <getting_started>` for more information.
"""

# uses https://pypi.org/project/sortedcontainers/
from sortedcontainers import SortedDict, SortedSet
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.plotting import register_matplotlib_converters
import pytz
from staircase.defaults import default
from staircase.docstrings.decorator import add_doc, append_doc
from staircase.docstrings import stairs_class as SC_docs
from staircase.core import ops, poly, stats
from staircase.core.tools.datetimes import (
    origin,
    _convert_date_to_float,
    _convert_float_to_date,
)

register_matplotlib_converters()


class Stairs:
    """
    An instance of a Stairs class is used to represent a :ref:`step function <getting_started.step_function>`.

    The Stairs class encapsulates a `SortedDict <http://www.grantjenks.com/docs/sortedcontainers/sorteddict.html>`_
    which is used to hold the points at which the step function changes, and by how much.

    See the :ref:`Stairs API <api.Stairs>` for details of methods.
    """

    def __init__(self, value=0, use_dates=default, tz=default):
        """
        Initialise a Stairs instance.

        Parameters
        ----------
        value : float, default 0
            The value of the step function at negative infinity.
        use_dates: bool, default False
            Allows the step function to be defined with `Pandas.Timestamp <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timestamp.html>`_.

        Returns
        -------
        :class:`Stairs`
        """
        if use_dates == default:
            use_dates = default.get_default("use_dates")
        if tz == default:
            tz = default.get_default("tz")

        self._sorted_dict = SortedDict()
        if isinstance(value, dict):
            self._sorted_dict = SortedDict(value)
        else:
            self._sorted_dict = SortedDict()
            self._sorted_dict[float("-inf")] = value
        self.use_dates = use_dates
        self.tz = tz
        self.cached_cumulative = None

        # runtime._add_operations(self, use_dates)

        self._get = self._sorted_dict.get
        self._items = self._sorted_dict.items
        self._keys = self._sorted_dict.keys
        self._sorted_dict_values = self._sorted_dict.values
        self._pop = self._sorted_dict.pop
        self._len = self._sorted_dict.__len__
        self._popitem = self._sorted_dict.popitem

    # DO NOT IMPLEMENT __len__ or __iter__, IT WILL CAUSE ISSUES WITH PANDAS SERIES PRETTY PRINTING

    def __getitem__(self, *args, **kwargs):
        """
        f'{dict.__getitem__.__doc__}'
        """
        return self._sorted_dict.__getitem__(*args, **kwargs)

    def __setitem__(self, key, value):
        """
        f'{dict.__setitem__.__doc__}'
        """
        self._sorted_dict.__setitem__(key, value)

    def copy(self, deep=None):
        """
        Returns a deep copy of this Stairs instance

        Parameters
        ----------
        deep : None
            Dummy parameter to keep pandas satisfied.

        Returns
        -------
        :class:`Stairs`
        """
        new_instance = Stairs(use_dates=self.use_dates, tz=self.tz)
        for key, value in self._items():
            new_instance[key] = value
        return new_instance

    @classmethod
    def from_cumulative(cls, cumulative, use_dates=False, tz=None):
        return cls(
            dict(
                zip(
                    cumulative.keys(),
                    np.insert(
                        np.diff(list(cumulative.values())),
                        0,
                        [next(iter(cumulative.values()))],
                    ),
                )
            ),
            use_dates,
            tz,
        )

    def plot(self, ax=None, **kwargs):
        """
        Makes a step plot representing the finite intervals belonging to the Stairs instance.

        Uses matplotlib as a backend.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`, default None
            Allows the axes, on which to plot, to be specified
        **kwargs
            Options to pass to :function: `matplotlib.pyplot.step`

        Returns
        -------
        :class:`matplotlib.axes.Axes`
        """
        if ax is None:
            _, ax = plt.subplots()

        cumulative = self._cumulative()
        if self.use_dates:
            register_matplotlib_converters()
            x = list(cumulative.keys())
            if len(x) > 1:
                x[0] = (
                    x[1] - 0.00001
                )  # first element would otherwise be -inf which can't be plotted
                ax.step(
                    _convert_float_to_date(x, self.tz),
                    list(cumulative.values()),
                    where="post",
                    **kwargs,
                )
        else:
            ax.step(cumulative.keys(), cumulative.values(), where="post", **kwargs)
        return ax

    def _cumulative(self):
        if self.cached_cumulative is None:
            self.cached_cumulative = SortedDict(
                zip(self._keys(), np.cumsum(self._sorted_dict_values()))
            )
        return self.cached_cumulative

    def _reduce(self):
        to_remove = [key for key, val in self._items()[1:] if val == 0]
        for key in to_remove:
            self._pop(key)
        return self

    def __bool__(self):
        """
        Return True if and only if step function has a value of 1 everywhere.

        Returns
        -------
        boolean
        """
        if self.number_of_steps() >= 2:
            return float((~self).integrate()) < 0.0000001
        return dict(self._sorted_dict) == {float("-inf"): 1}

    @append_doc(SC_docs.describe_example)
    def describe(
        self, lower=float("-inf"), upper=float("inf"), percentiles=(25, 50, 75)
    ):
        """
        Generate descriptive statistics.

        Parameters
        ----------
        lower : int, float or pandas.Timestamp
            lower bound of the interval on which to perform the calculation
        upper : int, float or pandas.Timestamp
            upper bound of the interval on which to perform the calculation
        percentiles: array-like of float, default [25, 50, 70]
            The percentiles to include in output.  Numbers should be in the range 0 to 100.

        Returns
        -------
        :class:`pandas.Series`

        See Also
        --------
        Stairs.mean, Stairs.std, Stairs.min, Stairs.percentile, Stairs.max
        """
        percentilestairs = self.percentile_stairs(lower, upper)
        return pd.Series(
            {
                **{
                    "unique": percentilestairs.clip(0, 100).number_of_steps() - 1,
                    "mean": self.mean(lower, upper),
                    "std": self.std(lower, upper),
                    "min": self.min(lower, upper),
                },
                **{f"{perc}%": percentilestairs(perc) for perc in percentiles},
                **{"max": self.max(lower, upper),},
            }
        )

    @append_doc(SC_docs.cov_example)
    def cov(self, other, lower=float("-inf"), upper=float("inf"), lag=0, clip="pre"):
        """
        Calculates either covariance, autocovariance or cross-covariance.

        The calculation is between two step functions described by *self* and *other*.
        If lag is None or 0 then covariance is calculated, otherwise cross-covariance is calculated.
        Autocovariance is a special case of cross-covariance when *other* is equal to *self*.

        Parameters
        ----------
        other: :class:`Stairs`
            the stairs instance with which to compute the covariance
        lower : int, float or pandas.Timestamp
            lower bound of the domain on which to perform the calculation
        upper : int, float or pandas.Timestamp
            upper bound of the domain on which to perform the calculation
        lag : int, float, pandas.Timedelta
            a pandas.Timedelta is only valid when using dates.
            If using dates and delta is an int or float, then it is interpreted as a number of hours.
        clip : {'pre', 'post'}, default 'pre'
            only relevant when lag is non-zero.  Determines if the domain is applied before or after *other* is translated.
            If 'pre' then the domain over which the calculation is performed is the overlap
            of the original domain and the translated domain.

        Returns
        -------
        float
            The covariance (or cross-covariance) between *self* and *other*

        See Also
        --------
        Stairs.corr, staircase.cov, staircase.corr
        """
        if lag != 0:
            assert clip in ["pre", "post"]
            if clip == "pre" and upper != float("inf"):
                upper = upper - lag
            other = other.shift(-lag)
        return (self * other).mean(lower, upper) - self.mean(lower, upper) * other.mean(
            lower, upper
        )

    @append_doc(SC_docs.corr_example)
    def corr(self, other, lower=float("-inf"), upper=float("inf"), lag=0, clip="pre"):
        """
        Calculates either correlation, autocorrelation or cross-correlation.

        All calculations are based off the `Pearson correlation coefficient <https://en.wikipedia.org/wiki/Pearson_correlation_coefficient>`_.

        The calculation is between two step functions described by *self* and *other*.
        If lag is None or 0 then correlation is calculated, otherwise cross-correlation is calculated.
        Autocorrelation is a special case of cross-correlation when *other* is equal to *self*.

        Parameters
        ----------
        other: :class:`Stairs`
            the stairs instance with which to compute the correlation
        lower : int, float or pandas.Timestamp
            lower bound of the interval on which to perform the calculation
        upper : int, float or pandas.Timestamp
            upper bound of the interval on which to perform the calculation
        lag : int, float, pandas.Timedelta
            a pandas.Timedelta is only valid when using dates.
            If using dates and delta is an int or float, then it is interpreted as a number of hours.
        clip : {'pre', 'post'}, default 'pre'
            only relevant when lag is non-zero.  Determines if the domain is applied before or after *other* is translated.
            If 'pre' then the domain over which the calculation is performed is the overlap
            of the original domain and the translated domain.

        Returns
        -------
        float
            The correlation (or cross-correlation) between *self* and *other*

        See Also
        --------
        Stairs.cov, staircase.corr, staircase.cov
        """
        if lag != 0:
            assert clip in ["pre", "post"]
            if clip == "pre" and upper != float("inf"):
                upper = upper - lag
            other = other.shift(-lag)
        return self.cov(other, lower, upper) / (
            self.std(lower, upper) * other.std(lower, upper)
        )

    @append_doc(SC_docs.min_example)
    def min(
        self,
        lower=float("-inf"),
        upper=float("inf"),
        lower_how="right",
        upper_how="left",
    ):
        """
        Calculates the minimum value of the step function

        If an interval which to calculate over is specified it is interpreted
        as a closed interval, with *lower_how* and *upper_how* indicating how the step function
        should be evaluated at the at the endpoints of the interval.

        Parameters
        ----------
        lower : int, float or pandas.Timestamp, optional
            lower bound of the interval on which to perform the calculation
        upper : int, float or pandas.Timestamp, optional
            upper bound of the interval on which to perform the calculation
        lower_how: {'left', 'right'}, default 'right'
            Determines how the step function should be evaluated at *lower*.
            If 'left' then :math:`\\lim_{x \\to lower^{-}} f(x)` is included in the calculation.
        upper_how: {'left', 'right'}, default 'left'
            Determines how the step function should be evaluated at *upper*.
            If 'right' then :math:`\\lim_{x \\to upper^{+}} f(x)` is included in the calculation.

        Returns
        -------
        float
            The minimum value of the step function

        See Also
        --------
        Stairs.max, staircase.min
        """
        return min(self.values_in_range(lower, upper, lower_how, upper_how))

    @append_doc(SC_docs.max_example)
    def max(
        self,
        lower=float("-inf"),
        upper=float("inf"),
        lower_how="right",
        upper_how="left",
    ):
        """
        Calculates the maximum value of the step function

        If an interval which to calculate over is specified it is interpreted
        as a closed interval, with *lower_how* and *upper_how* indicating how the step function
        should be evaluated at the at the endpoints of the interval.

        Parameters
        ----------
        lower : int, float or pandas.Timestamp, optional
            lower bound of the interval on which to perform the calculation
        upper : int, float or pandas.Timestamp, optional
            upper bound of the interval on which to perform the calculation
        lower_how: {'left', 'right'}, default 'right'
            Determines how the step function should be evaluated at *lower*.
            If 'left' then :math:`\\lim_{x \\to lower^{-}} f(x)` is included in the calculation.
        upper_how: {'left', 'right'}, default 'left'
            Determines how the step function should be evaluated at *upper*.
            If 'right' then :math:`\\lim_{x \\to upper^{+}} f(x)` is included in the calculation.

        Returns
        -------
        float
            The maximum value of the step function

        See Also
        --------
        Stairs.min, staircase.max
        """
        return max(self.values_in_range(lower, upper, lower_how, upper_how))

    @append_doc(SC_docs.shift_example)
    def shift(self, delta):
        """
        Returns a stairs instance corresponding to a horizontal translation by delta

        If delta is positive the corresponding step function is moved right.
        If delta is negative the corresponding step function is moved left.

        Parameters
        ----------
        delta : int, float or pandas.Timedelta
            the amount by which to translate.  A pandas.Timedelta is only valid when using dates.
            If using dates and delta is an int or float, then it is interpreted as a number of hours.

        Returns
        -------
        :class:`Stairs`

        See Also
        --------
        Stairs.diff
        """
        if isinstance(delta, pd.Timedelta):
            assert self.use_dates, "delta is of type pandas.Timedelta, expected float"
            delta = delta.total_seconds() / 3600
        return Stairs(
            dict(zip((key + delta for key in self._keys()), self._sorted_dict_values())),
            self.use_dates,
            self.tz,
        )

    @append_doc(SC_docs.diff_example)
    def diff(self, delta):
        """
        Returns a stairs instance corresponding to the difference between the step function corresponding to *self*
        and the same step-function translated by delta.

        Parameters
        ----------
        delta : int, float or pandas.Timedelta
            the amount by which to translate.  A pandas.Timestamp is only valid when using dates.
            If using dates and delta is an int or float, then it is interpreted as a number of hours.

        Returns
        -------
        :class:`Stairs`

        See Also
        --------
        Stairs.shift
        """
        return self - self.shift(delta)

    @append_doc(SC_docs.rolling_mean_example)
    def rolling_mean(self, window=(0, 0), lower=float("-inf"), upper=float("inf")):
        """
        Returns coordinates defining rolling mean

        The rolling mean of a step function is a continous piece-wise linear function, hence it can
        be described by a sequence of x,y coordinates which mark where function changes gradient.  These
        x,y coordinates are returned as a :class:`pandas.Series` which could then be used with
        :meth:`matplotlib.axes.Axes.plot`, or equivalent, to visualise.

        A rolling mean requires a window around a point x (referred to as the focal point) to be defined.
        In this implementation the window is defined by two values paired into an array-like parameter called *window*.
        These two numbers are the distance from the focal point to the left boundary of the window, and the right boundary
        of the window respectively.  This allows for trailing windows, leading windows and everything between
        (including a centred window).

        If *lower* or *upper* is specified then only coordinates corresponding to windows contained within
        [lower, upper] are included.

        Parameters
        ----------
        window : array-like of int, float or pandas.Timedelta
            should be length of 2. Defines distances from focal point to window boundaries.
        lower : int, float, pandas.Timestamp, or None, default None
            used to indicate the lower bound of the domain of the calculation
        upper : int, float, pandas.Timestamp, or None, default None
            used to indicate the upper bound of the domain of the calculation

        Returns
        -------
        :class:`pandas.Series`

        See Also
        --------
        Stairs.mean
        """
        assert len(window) == 2, "Window should be a listlike object of length 2."
        left_delta, right_delta = window
        clipped = self.clip(lower, upper)
        if clipped.use_dates:
            left_delta = pd.Timedelta(left_delta, "h")
            right_delta = pd.Timedelta(right_delta, "h")
        change_points = list(
            SortedSet(
                [c - right_delta for c in clipped.step_changes().keys()]
                + [c - left_delta for c in clipped.step_changes().keys()]
            )
        )
        s = pd.Series(
            clipped.sample(change_points, aggfunc="mean", window=window),
            index=change_points,
        )
        if lower != float("-inf"):
            s = s.loc[s.index >= lower - left_delta]
        if upper != float("inf"):
            s = s.loc[s.index <= upper - right_delta]
        return s

    def to_dataframe(self):
        """
        Returns a pandas.DataFrame with columns 'start', 'end' and 'value'

        The rows of the dataframe can be interpreted as the interval definitions
        which make up the step function.

        Returns
        -------
        :class:`pandas.DataFrame`
        """
        starts = self._keys()
        ends = self._keys()[1:]
        if self.use_dates:
            starts = [pd.NaT] + _convert_float_to_date(np.array(starts[1:]), self.tz)
            ends = _convert_float_to_date(np.array(ends), self.tz) + [pd.NaT]
        else:
            ends.append(float("inf"))
        values = self._cumulative().values()
        df = pd.DataFrame(
            {"start": list(starts), "end": list(ends), "value": list(values)}
        )  # bugfix for pandas 1.1
        return df

    @append_doc(SC_docs.number_of_steps_example)
    def number_of_steps(self):
        """
        Calculates the number of step changes

        Returns
        -------
        int

        See Also
        --------
        Stairs.step_changes
        """
        return len(self._keys()) - 1

    def __str__(self):
        """
        Return str(self)
        """
        tzinfo = f", tz={self.tz}" if self.use_dates else ""
        return f"<staircase.Stairs, id={id(self)}, dates={self.use_dates}{tzinfo}>"

    def __repr__(self):
        """
        Return string representation of Stairs
        """
        return str(self)

    def __call__(self, *args, **kwargs):
        return self.sample(*args, **kwargs)


poly.add_methods(Stairs)
ops.add_operations(Stairs)
stats.add_methods(Stairs)



sample_docstring = """
Evaluates the value of the step function at one, or more, points.

This method can be used to directly sample values of the corresponding step function at the points
provided, or alternatively calculate aggregations over some window around each point.  The first of these
is performed when *aggfunc* is None.

If *aggfunc* is None then the results of this function should be considered as :math:`\\lim_{x \\to z^{-}} f(x)`
or :math:`\\lim_{x \\to z^{+}} f(x)`, when how = 'left' or how = 'right' respectively. See
:ref:`A note on interval endpoints<getting_started.interval_endpoints>` for an explanation.

If *aggfunc* is not None then a window, around each point x (referred to as the focal point), over which to aggregate is required.
The window is defined by two values paired into an array-like parameter called *window*.
These two scalars are the distance from the focal point to the left boundary of the window, and the right boundary
of the window respectively.

The function can be called using parentheses.  See example below.

Parameters
----------
x : int, float or vector data
    Values at which to evaluate the function
how : {'left', 'right'}, default 'right'
    Only relevant if *aggfunc* is None.
    If points where step changes occur do not coincide with x then this parameter
    has no effect.  Where a step changes occurs at a point given by x, this parameter
    determines if the step function is evaluated at the interval to the left, or the right.
aggfunc: {'mean', 'median', 'mode', 'max', 'min', 'std', None}.  Default None.
    A string corresponding to the aggregating function
window : array-like of int, float or pandas.Timedelta, optional
    Only relevant if *aggfunc* is not None.  Should be length of 2. Defines distances from focal point to window boundaries.
lower_how: {'left', 'right'}, default 'right'
    Only relevant if *aggfunc* is not None.  Determines how the left window boundary should be evaluated.
    If 'left' then :math:`\\lim_{x \\to lower_how^{-}} f(x)` is included in the window.
upper_how: {'left', 'right'}, default 'left'
    Only relevant if *aggfunc* is not None.  Determines how the right window boundary should be evaluated.
    If 'right' then :math:`\\lim_{x \\to upper_how^{+}} f(x)` is included in the window.

Returns
-------
float, or list of floats

See Also
--------
staircase.sample

Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1(3.5)
    1
    >>> s1([1, 2, 4.5, 6])
    [1, 0, -1, 0]
    >>> s1([1, 2, 4.5, 6], how="left")
    [0, 1, -1, 0]
    >>> s1([1, 2, 4.5], aggfunc="mean", window=(-0.5, 0.5))
    [0.5, 0.5, -1.0]
    >>> s1([1, 2, 4.5], aggfunc="max", window=(-0.5, 0.5))
    [1, 1, -1]
    >>> s1([1, 2, 4.5], aggfunc="max", window=(-0.5, 0.5), lower_how="left")
    [1, 1, 1]
    >>> s1([1, 2, 4.5], aggfunc="max", window=(-0.5, 0.5), upper_how="right")
    [1, 1, 0]
"""

resample_docstring = """
Evaluates the value of the step function at one, or more, points and
creates a new Stairs instance whose step changes occur at a subset of these
points.  The new instance and self have the same values when evaluated at x.

Parameters
----------
x : int, float or vector data
    Values at which to evaluate the function
how : {'left', 'right'}, default 'right'
    If points where step changes occur do not coincide with x then this parameter
    has no effect.  Where a step changes occurs at a point given by x, this parameter
    determines if the step function is evaluated at the interval to the left, or the right.
aggfunc: {'mean', 'median', 'mode', 'max', 'min', 'std', None}.  Default None.
    A string corresponding to the aggregating function
window : array-like of int, float or pandas.Timedelta, optional
    Only relevant if *aggfunc* is not None.  Should be length of 2. Defines distances from focal point to window boundaries.
lower_how: {'left', 'right'}, default 'right'
    Only relevant if *aggfunc* is not None.  Determines how the left window boundary should be evaluated.
    If 'left' then :math:`\\lim_{x \\to lower_how^{-}} f(x)` is included in the window.
upper_how: {'left', 'right'}, default 'left'
    Only relevant if *aggfunc* is not None.  Determines how the right window boundary should be evaluated.
    If 'right' then :math:`\\lim_{x \\to upper_how^{+}} f(x)` is included in the window.

Returns
-------
:class:`Stairs`

See Also
--------
staircase.resample

Examples
--------

.. plot::
    :context: close-figs

    >>> stair_list = [s1, s1.resample([1.5,2.5,4,4.5])]
    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5), sharey=True, sharex=True)
    >>> for ax, title, stair_instance in zip(axes, ("s1", "s1 resampled"), stair_list):
    ...     stair_instance.plot(ax)
    ...     ax.set_title(title)
"""

layer_docstring = """
Changes the value of the step function.


Parameters
----------
start : int, float or vector data, optional
    start time(s) of the interval(s)
end : int, float or vector data, optional
    end time(s) of the interval(s)
value: int, float or vector data, optional
    value(s) of the interval(s)

Returns
-------
:class:`Stairs`
    The current instance is returned to facilitate method chaining

Examples
--------

.. plot::
    :context: close-figs

    >>> import staircase as sc
    ... (sc.Stairs()
    ...     .layer(1,3)
    ...     .layer(4,5,-2)
    ...     .plot()
    ... )

.. plot::
    :context: close-figs

    >>> import pandas as pd
    >>> import staircase as sc
    >>> data = pd.DataFrame({"starts":[1,4,5.5],
    ...                      "ends":[3,5,7],
    ...                      "values":[-1,2,-3]})
    >>> data
       starts  ends  values
    0     1.0     3      -1
    1     4.0     5       2
    2     5.5     7      -3
    >>> (sc.Stairs(1.5)
    ...     .layer(data["starts"], data["ends"], data["values"])
    ...     .plot()
    ... )
"""

step_changes_docstring = """
Returns a dictionary of key, value pairs of indicating where step changes occur in the step function, and the change in value

Returns
-------
dictionary

See Also
--------
Stairs.number_of_steps

Examples
--------

.. plot::
    :context: close-figs

    >>> s1.plot()
    >>> s1.step_changes()
    {1: 1, 2: -1, 3: 1, 4: -2, 5: 1}
"""


values_in_range_docstring = """
Returns the range of the step function as a set of discrete values.

Parameters
----------
lower : int, float or pandas.Timestamp, optional
    lower bound of the interval on which to perform the calculation
upper : int, float or pandas.Timestamp, optional
    upper bound of the interval on which to perform the calculation
lower_how: {'left', 'right'}, default 'right'
    Determines how the step function should be evaluated at *lower*.
    If 'left' then :math:`\\lim_{x \\to lower^{-}} f(x)` is included in the calculation.
upper_how: {'left', 'right'}, default 'left'
    Determines how the step function should be evaluated at *upper*.
    If 'right' then :math:`\\lim_{x \\to upper^{+}} f(x)` is included in the calculation.

Returns
-------
set of floats

Examples
--------

.. plot::
    :context: close-figs

    >>> s2.plot()
    >>> s2.values_in_range()
    {-1.0, 0.0, 0.5}
    >>> s2.values_in_range(lower=2)
    {-1.0, 0.0}
    >>> s2.values_in_range(lower=2, lower_how="left")
    {-1.0, 0.0, 0.5}
    >>> s2.values_in_range(upper=2)
    {0.0, 0.5}
    >>> s2.values_in_range(upper=3, upper_how="right")
    {-1.0, 0.0, 0.5}

"""

clip_docstring = """
Returns a copy of *self* which is zero-valued everywhere outside of [lower, upper)

Parameters
----------
lower : int, float or pandas.Timestamp
    lower bound of the interval
upper : int, float or pandas.Timestamp
    upper bound of the interval

Returns
-------
:class:`Stairs`
    Returns a copy of *self* which is zero-valued everywhere outside of [lower, upper)

Examples
--------

.. plot::
    :context: close-figs

    >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5))
    >>> s1.plot(axes[0])
    >>> s1.clip(2,4).plot(axes[1])
    >>> s1.clip(2,4).mean()
    0.5

"""



import pandas as pd
import numpy as np
from staircase.core.tools.datetimes import (
    _convert_date_to_float,
    _convert_float_to_date,
    _using_dates,
)
from staircase.core.tools import _get_union_of_points
from staircase.util._decorators import Appender
from staircase.core.collections import docstrings
from staircase.core.stairs import Stairs
from staircase.core.poly.methods import _sample_raw


@Appender(docstrings.sample_example, join="\n", indents=1)
def sample(collection, points=None, how="right", expand_key=True):
    """
    Takes a dict-like collection of Stairs instances and evaluates their values across a common set of points.

    Technically the results of this function should be considered as :math:`\\lim_{x \\to z^{-}} f(x)`
    or :math:`\\lim_{x \\to z^{+}} f(x)`, when how = 'left' or how = 'right' respectively. See
    :ref:`A note on interval endpoints<getting_started.interval_endpoints>` for an explanation.


    Parameters
    ----------
    collection : dictionary or pandas.Series
        The Stairs instances at which to evaluate
    points : int, float or vector data
        The points at which to sample the Stairs instances
    how : {'left', 'right'}, default 'right'
        if points where step changes occur do not coincide with x then this parameter
        has no effect.  Where a step changes occurs at a point given by x, this parameter
        determines if the step function is evaluated at the interval to the left, or the right.
    expand_key: boolean, default True
        used when collection is a multi-index pandas.Series.  Indicates if index should be expanded from
        tuple to columns in a dataframe.

    Returns
    -------
    :class:`pandas.DataFrame`
        A dataframe, in tidy format, with three columns: points, key, value.  The column key contains
        the identifiers used in the dict-like object specified by 'collection'.

    See Also
    --------
    Stairs.sample
    """
    use_dates, tz = _using_dates(collection)
    # assert len(set([type(x) for x in collection.values()])) == 1, "collection must contain values of same type"
    if points is None:
        points = _get_union_of_points(collection)
        points.discard(float("-inf"))
        if use_dates:
            points = _convert_float_to_date(
                list(points), tz
            )  # bugfix - pandas>=1.1 breaks with points as type SortedSet
    else:
        if not hasattr(points, "__iter__"):
            points = [points]
    points = list(points)  # bugfix - pandas>=1.1 breaks with points as type SortedSet
    result = pd.DataFrame(
        {
            "points": points,
            **{
                key: stairs.sample(points, how=how)
                for key, stairs in collection.items()
            },
        }
    ).melt(id_vars="points", var_name="key")
    if (
        isinstance(collection, pd.Series)
        and expand_key
        and len(collection.index.names) > 1
    ):
        try:
            result = result.join(
                pd.DataFrame(result.key.tolist(), columns=collection.index.names)
            ).drop(columns="key")
        except Exception:
            pass
    return result


@Appender(docstrings.aggregate_example, join="\n", indents=1)
def aggregate(collection, func, points=None):
    """
    Takes a collection of Stairs instances and returns a single instance representing the aggregation.

    Parameters
    ----------
    collection: tuple, list, numpy array, dict or pandas.Series
        The Stairs instances to aggregate
    func: a function taking a 1 dimensional vector of floats, and returning a single float
        The function to apply, eg numpy.max
    points: vector of floats or dates
        Points at which to evaluate.  Defaults to union of all step changes.  Equivalent to applying Stairs.resample().

    Returns
    ----------
    :class:`Stairs`

    Notes
    -----
    The points at which to aggregate will include -infinity whether explicitly included or not.

    See Also
    --------
    staircase.mean, staircase.median, staircase.min, staircase.max
    """
    if isinstance(collection, dict) or isinstance(collection, pd.Series):
        Stairs_dict = collection
    else:
        Stairs_dict = dict(enumerate(collection))
    use_dates, tz = _using_dates(collection)
    df = sample(Stairs_dict, points, expand_key=False)
    aggregation = df.pivot_table(
        index="points", columns="key", values="value"
    ).aggregate(func, axis=1)
    if use_dates:
        aggregation.index = _convert_date_to_float(aggregation.index, tz=tz)
    aggregation[float("-inf")] = func(
        [_sample_raw(val, float("-inf")) for key, val in Stairs_dict.items()]
    )
    step_changes = aggregation.sort_index().diff().fillna(0)
    step_changes[float("-inf")] = aggregation[float("-inf")]
    # groupby.sum is necessary on next line as step_changes series may not have unique index elements
    return Stairs(
        dict(step_changes.groupby(level=0).sum()), use_dates=use_dates, tz=tz
    )._reduce()


@Appender(docstrings.mean_docstring, join="\n", indents=1)
def _mean(collection):
    return aggregate(collection, np.mean)


@Appender(docstrings.median_docstring, join="\n", indents=1)
def _median(collection):
    return aggregate(collection, np.median)


@Appender(docstrings.min_docstring, join="\n", indents=1)
def _min(collection):
    return aggregate(collection, np.min)


@Appender(docstrings.max_docstring, join="\n", indents=1)
def _max(collection):
    return aggregate(collection, np.max)


def resample(container, x, how="right"):
    """
    Applies the Stairs.resample function to a 1D container, eg tuple, list, numpy array, pandas series, dictionary

    Returns
    -------
    type(container)

    See Also
    --------
    Stairs.resample
    """
    if isinstance(container, dict):
        return {key: s.resample(x, how) for key, s in container}
    if isinstance(container, pd.Series):
        return pd.Series(
            [s.resample(x, how) for s in container.values], index=container.index
        )
    if isinstance(container, np.ndarray):
        return np.array([s.resample(x, how) for s in container])
    return type(container)([s.resample(x, how) for s in container])


def _make_corr_cov_func(docstring, stairs_method, assume_ones_diagonal):
    @Appender(docstring, join="\n", indents=1)
    def func(collection, lower=float("-inf"), upper=float("inf")):
        series = pd.Series(collection)
        size = series.shape[0]
        vals = np.ones(shape=(size, size))
        for i in range(size):
            for j in range(i + assume_ones_diagonal, size):
                vals[i, j] = stairs_method(series.iloc[i], series.iloc[j], lower, upper)
                vals[j, i] = vals[i, j]
        return pd.DataFrame(vals, index=series.index, columns=series.index)

    return func


corr = _make_corr_cov_func(
    docstrings.corr_docstring, Stairs.corr, assume_ones_diagonal=True
)
cov = _make_corr_cov_func(
    docstrings.cov_docstring, Stairs.cov, assume_ones_diagonal=False
)



from staircase.core.poly.methods import (
    _layer_multiple,
    clip,
    get_integral_and_mean,
    layer,
    resample,
    sample,
    step_changes,
    values_in_range,
)


def add_methods(cls):
    cls._layer_multiple = _layer_multiple
    cls.clip = clip
    cls.get_integral_and_mean = get_integral_and_mean
    cls.layer = layer
    cls.resample = resample
    cls.sample = sample
    cls.step_changes = step_changes
    cls.values_in_range = values_in_range



from staircase.core.stairs import Stairs

from staircase.core.collections.functions import (
    aggregate,
    corr,
    cov,
    _max as max,
    _mean as mean,
    _median as median,
    _min as min,
    resample,
    sample,
)

from staircase.defaults import default

from staircase.core.stats import hist_from_ecdf

from .test_data import make_test_data


def get_version():
    def get_version_post_py38():
        from importlib.metadata import version

        return version(__name__)

    def get_version_pre_py38():
        from pkg_resources import get_distribution

        return get_distribution(__name__).version

    def default_version():
        return "unknown"

    for func in (get_version_post_py38, get_version_pre_py38, default_version):
        try:
            return func()
        except Exception:
            pass


__version__ = get_version()



import numpy as np
from pandas import Timestamp
from staircase.core.stats.distribution import percentile_stairs
from staircase.core.tools.datetimes import _convert_date_to_float
from staircase.util._decorators import Appender
from staircase.core.stats import docstrings
import staircase as sc


@Appender(docstrings.integrate_docstring, join="\n", indents=1)
def integrate(self, lower=float("-inf"), upper=float("inf")):
    area, _ = self.get_integral_and_mean(lower, upper)
    return area


@Appender(docstrings.mean_docstring, join="\n", indents=1)
def mean(self, lower=float("-inf"), upper=float("inf")):
    _, mean = self.get_integral_and_mean(lower, upper)
    return mean


@Appender(docstrings.percentile_docstring, join="\n", indents=1)
def percentile(self, x, lower=float("-inf"), upper=float("inf")):
    assert 0 <= x <= 100
    percentiles = percentile_stairs(self, lower, upper)
    return (percentiles(x, how="left") + percentiles(x, how="right")) / 2


@Appender(docstrings.median_docstring, join="\n", indents=1)
def median(self, lower=float("-inf"), upper=float("inf")):
    return percentile(self, 50, lower, upper)


@Appender(docstrings.mode_docstring, join="\n", indents=1)
def mode(self, lower=float("-inf"), upper=float("inf")):
    df = (
        self.clip(lower, upper)
        .to_dataframe()
        .iloc[1:-1]
        .assign(duration=lambda df: df.end - df.start)
    )
    return df.value.loc[df.duration.idxmax()]


@Appender(docstrings.var_docstring, join="\n", indents=1)
def var(self, lower=float("-inf"), upper=float("inf")):
    percentile_minus_mean = (
        percentile_stairs(self, lower, upper) - self.mean(lower, upper)
    )._cumulative()
    return (
        sc.Stairs.from_cumulative(
            dict(
                zip(
                    percentile_minus_mean.keys(),
                    (val * val for val in percentile_minus_mean.values()),
                )
            )
        ).integrate(0, 100)
        / 100
    )


@Appender(docstrings.std_docstring, join="\n", indents=1)
def std(self, lower=float("-inf"), upper=float("inf")):
    return np.sqrt(var(self, lower, upper))



import pandas as pd
import numpy as np
from sortedcontainers import SortedDict
from staircase.util._decorators import Appender
from staircase.core.poly import docstrings
from staircase.core.stats.docstrings import integral_and_mean_docstring
import staircase as sc
from staircase.core.tools import (
    _from_cumulative,
    _get_stairs_method,
    _verify_window,
)
from staircase.core.tools.datetimes import (
    _convert_date_to_float,
    _convert_float_to_date,
    _maybe_convert_from_timedeltas,
    _maybe_convert_from_timestamps,
)


def _sample_raw(self, x, how="right"):
    """
    Evaluates the value of the step function at one, or more, points.

    Technically the results of this function should be considered as :math:`\\lim_{x \\to z^{-}} f(x)`
    or :math:`\\lim_{x \\to z^{+}} f(x)`, when how = 'left' or how = 'right' respectively. See
    :ref:`A note on interval endpoints<getting_started.interval_endpoints>` for an explanation.

    Parameters
    ----------
    x : int, float or vector data
        values at which to evaluate the function
    how : {'left', 'right'}, default 'right'
        if points where step changes occur do not coincide with x then this parameter
        has no effect.  Where a step changes occurs at a point given by x, this parameter
        determines if the step function is evaluated at the interval to the left, or the right.

    Returns
    -------
    float, or list of floats

    See Also
    --------
    staircase.sample
    """
    assert how in ("right", "left")
    if hasattr(x, "__iter__"):
        new_instance = _layer_multiple(self.copy(), x, None, [0] * len(x))
        cumulative = new_instance._cumulative()
        if how == "right":
            return [cumulative[_x] for _x in x]
        else:
            shifted_cumulative = SortedDict(
                zip(cumulative.keys()[1:], cumulative.values()[:-1])
            )
            if float("-inf") in x:
                vals = [self[float("-inf")]]
            else:
                vals = []
            vals.extend([val for key, val in shifted_cumulative.items() if key in x])
            return vals
    elif x == float("-inf"):
        return self._sorted_dict_values()[0]
    else:
        cumulative = self._cumulative()
        if how == "right":
            preceding_boundary_index = cumulative.bisect_right(x) - 1
        else:
            preceding_boundary_index = cumulative.bisect_left(x) - 1
        return cumulative.values()[preceding_boundary_index]


def _sample_agg(self, x, window, aggfunc, lower_how="right", upper_how="left"):
    """
    Evaluates the aggregation of the step function over a window around one, or more, points.

    The window around each point is defined by two values paired into an array-like parameter called *window*.
    These two scalars are the distance from the point to the left boundary of the window, and the right boundary
    of the window respectively.


    Parameters
    ----------
    x : int, float or vector data
        values at which to evaluate the function
    aggfunc: {'mean', 'median', 'mode', 'max', 'min', 'std'}
        A string corresponding to the aggregating function
    window : array-like of int, float, optional
        Should be length of 2. Defines distances from focal point to window boundaries.
    lower_how: {'left', 'right'}, default 'right'
        Determines how the left window boundary should be evaluated.
        If 'left' then :math:`\\lim_{x \\to lower_how^{-}} f(x)` is included in the window.
    upper_how: {'left', 'right'}, default 'left'
        Determines how the right window boundary should be evaluated.
        If 'right' then :math:`\\lim_{x \\to upper_how^{+}} f(x)` is included in the window.

    Returns
    -------
    float, or list of floats

    See Also
    --------
    staircase.sample
    """
    assert len(window) == 2, "Window should be a array-like object of length 2."
    if isinstance(aggfunc, str):
        aggfunc = _get_stairs_method(aggfunc)
    left_delta, right_delta = window
    _verify_window(left_delta, right_delta, 0)
    kwargs = (
        {"lower_how": lower_how, "upper_how": upper_how}
        if aggfunc in [sc.Stairs.min, sc.Stairs.max]
        else {}
    )
    if not hasattr(x, "__iter__"):
        return aggfunc(self, lower=x + left_delta, upper=x + right_delta, **kwargs)
    return [
        aggfunc(self, lower=point + left_delta, upper=point + right_delta, **kwargs)
        for point in x
    ]


@Appender(docstrings.sample_docstring, join="\n", indents=1)
def _sample(
    self,
    x,
    how="right",
    aggfunc=None,
    window=(0, 0),
    lower_how="right",
    upper_how="left",
):
    # not using dates
    if aggfunc is None:
        return _sample_raw(self, x, how)
    else:
        return _sample_agg(self, x, window, aggfunc, lower_how, upper_how)


@Appender(docstrings.sample_docstring, join="\n", indents=1)
def sample(
    self,
    x,
    how="right",
    aggfunc=None,
    window=(0, 0),
    lower_how="right",
    upper_how="left",
):
    if self.use_dates:
        x = _convert_date_to_float(x, self.tz)
        window = _maybe_convert_from_timedeltas(window)
    return _sample(self, x, how, aggfunc, window, lower_how, upper_how)


@Appender(docstrings.resample_docstring, join="\n", indents=1)
def _resample(
    self,
    x,
    how="right",
    aggfunc=None,
    window=(0, 0),
    lower_how="right",
    upper_how="left",
):
    if aggfunc is not None:
        assert len(window) == 2, "Window should be a array-like object of length 2."
    if not hasattr(x, "__iter__"):
        x = [
            x,
        ]
    new_cumulative = SortedDict({float("-inf"): _sample(self, float("-inf"))})
    new_cumulative.update(
        {
            point: _sample(self, point, how, aggfunc, window, lower_how, upper_how)
            for point in x
        }
    )
    return _from_cumulative(new_cumulative, self.use_dates, self.tz)


@Appender(docstrings.resample_docstring, join="\n", indents=1)
def resample(
    self,
    x,
    how="right",
    aggfunc=None,
    window=(0, 0),
    lower_how="right",
    upper_how="left",
):
    if self.use_dates:
        x = _convert_date_to_float(x, self.tz)
        window = _maybe_convert_from_timedeltas(window)
    return _resample(self, x, how, aggfunc, window, lower_how, upper_how)


@Appender(docstrings.layer_docstring, join="\n", indents=1)
def _layer(self, start=None, end=None, value=None):
    if hasattr(start, "__iter__") or hasattr(end, "__iter__"):
        layer_func = _layer_multiple
    else:
        layer_func = _layer_single
    return layer_func(self, start, end, value)


@Appender(docstrings.layer_docstring, join="\n", indents=1)
def layer(self, start=None, end=None, value=None):
    if self.use_dates:
        start, end = _maybe_convert_from_timestamps((start, end), self.tz)
    return _layer(self, start, end, value)


def _layer_single(self, start=None, end=None, value=None):
    """
    Implementation of the layer function for when start parameter is single-valued
    """
    if pd.isna(start):
        start = float("-inf")
    if pd.isna(value):
        value = 1
    self[start] = self._get(start, 0) + value
    if start != float("-inf") and self[start] == 0:
        self._pop(start)

    if not pd.isna(end):
        self[end] = self._get(end, 0) - value
        if self[end] == 0 or end == float("inf"):
            self._pop(end)

    self.cached_cumulative = None
    return self


def _layer_multiple(self, starts=None, ends=None, values=None):
    """
    Implementation of the layer function for when start parameter is vector data
    """
    for vector in (starts, ends):
        if vector is not None and values is not None:
            assert len(vector) == len(values)

    if starts is None:
        starts = [float("-inf")] * len(ends)
    if ends is None:
        ends = []
    if values is None:
        values = [1] * max(len(starts), len(ends))

    for start, value in zip(starts, values):
        if pd.isna(start):
            start = float("-inf")
        self[start] = self._get(start, 0) + value
    for end, value in zip(ends, values):
        if not pd.isna(end):
            self[end] = self._get(end, 0) - value
    self.cached_cumulative = None
    return self


@Appender(docstrings.step_changes_docstring, join="\n", indents=1)
def step_changes(self):
    if self.use_dates:
        return dict(
            zip(_convert_float_to_date(self._keys()[1:], self.tz), self._sorted_dict_values()[1:])
        )
    return dict(self._items()[1:])


@Appender(docstrings.values_in_range_docstring, join="\n", indents=1)
def values_in_range(
    self, lower=float("-inf"), upper=float("inf"), lower_how="right", upper_how="left",
):
    if self.use_dates:
        lower, upper = _maybe_convert_from_timestamps((lower, upper), self.tz)
    return _values_in_range(self, lower, upper, lower_how, upper_how)


@Appender(docstrings.values_in_range_docstring, join="\n", indents=1)
def _values_in_range(
    self, lower=float("-inf"), upper=float("inf"), lower_how="right", upper_how="left",
):
    interior_points = [key for key in self._keys() if lower < key < upper]
    endpoint_vals = _sample_raw(self, [lower], how="right") + _sample_raw(
        self, [upper], how="left"
    )
    if lower_how == "left":
        endpoint_vals += _sample_raw(self, [lower], how="left")
    if upper_how == "right":
        endpoint_vals += _sample_raw(self, [upper], how="right")
    return set(_sample_raw(self, interior_points) + endpoint_vals)


@Appender(docstrings.clip_docstring, join="\n", indents=1)
def _clip(self, lower=float("-inf"), upper=float("inf")):
    assert (
        lower is not None and upper is not None
    ), "clip function should not be called with no parameters."
    assert (
        lower < upper
    ), "Value of parameter 'lower' must be less than the value of parameter 'upper'"
    cumulative = self._cumulative()
    left_boundary_index = cumulative.bisect_right(lower) - 1
    right_boundary_index = cumulative.bisect_right(upper) - 1
    value_at_left = cumulative.values()[left_boundary_index]
    value_at_right = cumulative.values()[right_boundary_index]
    s = dict(self._items()[left_boundary_index + 1 : right_boundary_index + 1])
    s[float("-inf")] = 0
    if lower != float("-inf"):
        s[float("-inf")] = 0
        s[lower] = value_at_left
    else:
        s[float("-inf")] = self[float("-inf")]
    if upper != float("inf"):
        s[upper] = s.get(upper, 0) - value_at_right
    return sc.Stairs(s, self.use_dates, self.tz)


@Appender(docstrings.clip_docstring, join="\n", indents=1)
def clip(self, lower=float("-inf"), upper=float("inf")):
    if self.use_dates:
        lower, upper = _maybe_convert_from_timestamps((lower, upper), self.tz)
    return _clip(self, lower, upper)


@Appender(integral_and_mean_docstring, join="\n", indents=1)
def _get_integral_and_mean(self, lower=float("-inf"), upper=float("inf")):
    new_instance = self.clip(lower, upper)
    if new_instance.number_of_steps() < 2:
        return 0, np.nan
    if lower != float("-inf"):
        new_instance[lower] = new_instance._get(lower, 0)
    if upper != float("inf"):
        new_instance[upper] = new_instance._get(upper, 0)
    cumulative = new_instance._cumulative()
    widths = np.subtract(cumulative.keys()[2:], cumulative.keys()[1:-1])
    heights = cumulative.values()[1:-1]
    area = np.multiply(widths, heights).sum()
    mean = area / (cumulative.keys()[-1] - cumulative.keys()[1])
    return area, mean


@Appender(integral_and_mean_docstring, join="\n", indents=1)
def get_integral_and_mean(self, lower=float("-inf"), upper=float("inf")):
    if self.use_dates:
        lower, upper = _maybe_convert_from_timestamps((lower, upper), self.tz)
    return _get_integral_and_mean(self, lower, upper)



import numpy as np
from staircase.core.tools import _sanitize_binary_operands
from staircase.util._decorators import Appender
from staircase.core.ops import docstrings
import staircase as sc


def _compare(cumulative, zero_comparator, use_dates=False, tz=None):
    truth = cumulative.copy()
    for key, value in truth.items():
        new_val = int(zero_comparator(float(value)))
        truth[key] = new_val
    deltas = [truth.values()[0]]
    deltas.extend(np.subtract(truth.values()[1:], truth.values()[:-1]))
    new_instance = sc.Stairs(dict(zip(truth.keys(), deltas)), use_dates, tz)
    new_instance._reduce()
    return new_instance


def _make_relational_func(comparator, docstring):
    @Appender(docstring, join="\n", indents=1)
    def func(self, other):
        self, other = _sanitize_binary_operands(self, other)
        return _compare(
            (other - self)._cumulative(),
            comparator,
            self.use_dates or other.use_dates,
            self.tz,
        )

    return func


lt = _make_relational_func(float(0).__lt__, docstrings.lt_docstring)
gt = _make_relational_func(float(0).__gt__, docstrings.gt_docstring)
le = _make_relational_func(float(0).__le__, docstrings.le_docstring)
ge = _make_relational_func(float(0).__ge__, docstrings.ge_docstring)
eq = _make_relational_func(float(0).__eq__, docstrings.eq_docstring)
ne = _make_relational_func(float(0).__ne__, docstrings.ne_docstring)


@Appender(docstrings.identical_docstring, join="\n", indents=1)
def identical(self, other):
    self, other = _sanitize_binary_operands(self, other)
    return bool(self == other)



sample_footer = """
    >>> sc.sample({"s1":s1, "s2":s2}, [1, 1.5, 2.5, 4])
       points   key   value
    0     1.0    s1     1.0
    1     1.5    s1     1.0
    2     2.5    s1     0.0
    3     4.0    s1    -1.0
    4     1.0    s2     0.5
    5     1.5    s2     0.5
    6     2.5    s2     0.0
    7     4.0    s2    -1.0
"""

_plot_titles_map = {
    "sample": ["s1", "s2"],
    "min": ["s1", "s2", "sc.min([s1,s2])"],
    "max": ["s1", "s2", "sc.max([s1,s2])"],
    "mean": ["s1", "s2", "sc.mean([s1,s2])"],
    "median": ["s1", "s2", "sc.median([s1,s2])"],
    "aggregate": ["s1", "s2", "sc.aggregate([s1,s2], np.std)"],
}

_example = """
Examples
--------

.. plot::
    :context: close-figs

    >>> stair_list = [{plots}]
    >>> fig, axes = plt.subplots(nrows=1, ncols={ncols}, figsize=(17,5), sharey=True, sharex=True)
    >>> for ax, title, stair_instance in zip(axes, ({plot_titles}), stair_list):
    ...     stair_instance.plot(ax)
    ...     ax.set_title(title)
"""


def _gen_example(operation):
    plot_titles = _plot_titles_map[operation]
    ncols = len(plot_titles)
    plots = ", ".join(plot_titles)
    return _example.format(plots=plots, ncols=ncols, plot_titles=plot_titles)


sample_example = _gen_example("sample") + sample_footer
min_example = _gen_example("min")
max_example = _gen_example("max")
mean_example = _gen_example("mean")
median_example = _gen_example("median")
aggregate_example = _gen_example("aggregate")

_corr_cov_header = """
Examples
--------

.. plot::
    :context: close-figs

    >>> import staircase as sc
    >>> pd.Series([s1, s2, s1+s2])
    0    <staircase.Stairs, id=2452772382088, dates=False>
    1    <staircase.Stairs, id=2452772381320, dates=False>
    2    <staircase.Stairs, id=2452772893512, dates=False>
    dtype: object
"""

cov_example = (
    _corr_cov_header
    + """
    >>> sc.cov(pd.Series([s1, s2, s1+s2], index=['s1', 's2', 's1+s2']))
                  s1         s2      s1+s2
    s1      0.687500   0.140496   0.652893
    s2      0.140496   0.471074   0.611570
    s1+ s2  0.652893   0.611570   1.264463

    >>> sc.cov([s1, s2, s1+s2])
              0          1          2
    0  0.687500   0.140496   0.652893
    1  0.140496   0.471074   0.611570
    2  0.652893   0.611570   1.264463
"""
)

corr_example = (
    _corr_cov_header
    + """
    >>> sc.corr(pd.Series([s1, s2, s1+s2], index=['s1', 's2', 's1+s2']))
                  s1         s2      s1+s2
    s1      1.000000   0.246878   0.700249
    s2      0.246878   1.000000   0.792407
    s1+ s2  0.700249   0.792407   1.000000

    >>> sc.corr([s1, s2, s1+s2])
              0          1          2
    0  1.000000   0.246878   0.700249
    1  0.246878   1.000000   0.792407
    2  0.700249   0.792407   1.000000
"""
)


def _get_example(calculation):

    return {
        "min": min_example,
        "max": max_example,
        "mean": mean_example,
        "median": median_example,
        "corr": corr_example,
        "cov": cov_example,
    }[calculation]


_see_also_map = {
    "min": "staircase.aggregate, staircase.mean, staircase.median, staircase.max",
    "max": "staircase.aggregate, staircase.mean, staircase.median, staircase.min",
    "mean": "staircase.aggregate, staircase.median, staircase.min, staircase.max",
    "median": "staircase.aggregate, staircase.mean, staircase.min, staircase.max",
    "corr": "Stairs.corr, staircase.cov",
    "cov": "Stairs.cov, staircase.corr",
}

_calc_map = {
    "min": "minimum",
    "max": "maximum",
    "corr": "correlation",
    "cov": "covariance",
}


_return_stairs_docstring = """
Takes a collection of Stairs instances and returns the {calc} of the corresponding step functions.

Parameters
----------
collection : tuple, list, numpy array, dict or pandas.Series
    The Stairs instances to aggregate using a {calc} function

Returns
-------
:class:`Stairs`

See Also
--------
{see_also}

{example}
"""

_corr_cov_docstring = """
Calculates the {calc} matrix for a collection of :class:`Stairs` instances

Parameters
----------
collection: :class:`pandas.Series`, dict, or array-like of :class:`Stairs` values
    the stairs instances with which to compute the {calc} matrix
lower : int, float or pandas.Timestamp
    lower bound of the interval on which to perform the calculation
upper : int, float or pandas.Timestamp
    upper bound of the interval on which to perform the calculation

Returns
-------
:class:`pandas.DataFrame`
    The {calc} matrix

See Also
--------
{see_also}

{example}
"""


def _gen_docstring(calculation):

    calc = _calc_map.get(calculation, calculation)
    see_also = _see_also_map[calculation]
    example = _get_example(calculation)

    if calculation in ("cov", "corr"):
        docstring = _corr_cov_docstring
    else:
        docstring = _return_stairs_docstring
    return docstring.format(calc=calc, see_also=see_also, example=example,)


min_docstring = _gen_docstring("min")
max_docstring = _gen_docstring("max")
mean_docstring = _gen_docstring("mean")
median_docstring = _gen_docstring("median")
cov_docstring = _gen_docstring("cov")
corr_docstring = _gen_docstring("corr")



import numpy as np
from sortedcontainers import SortedSet
from staircase.core.tools.datetimes import check_binop_timezones
import staircase as sc


def _sanitize_binary_operands(self, other, copy_other=False):
    if not isinstance(other, sc.Stairs):
        other = sc.Stairs(other, self.use_dates, self.tz)
    else:
        check_binop_timezones(self, other)
        if copy_other:
            other = other.copy()
    return self.copy(), other


def _get_union_of_points(collection):
    def dict_common_points():
        return collection.values()

    def series_common_points():
        return collection.values

    def array_common_points():
        return collection

    for func in (dict_common_points, series_common_points, array_common_points):
        try:
            stairs_instances = func()
            points = []
            for stair_instance in stairs_instances:
                points += list(stair_instance._keys())
            return SortedSet(points)
        except (AttributeError, TypeError):
            pass
    raise TypeError(
        "Collection should be a tuple, list, numpy array, dict or pandas.Series."
    )


def _get_stairs_method(name):
    return {
        "mean": sc.Stairs.mean,
        "median": sc.Stairs.median,
        "mode": sc.Stairs.mode,
        "max": sc.Stairs.max,
        "min": sc.Stairs.min,
    }[name]


def _verify_window(left_delta, right_delta, zero):
    assert left_delta <= zero, "left_delta must not be positive"
    assert right_delta >= zero, "right_delta must not be negative"
    assert right_delta - left_delta > zero, "window length must be non-zero"


def _from_cumulative(cumulative, use_dates=False, tz=None):
    return sc.Stairs(
        dict(
            zip(
                cumulative.keys(),
                np.insert(
                    np.diff(list(cumulative.values())),
                    0,
                    [next(iter(cumulative.values()))],
                ),
            )
        ),
        use_dates,
        tz,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['staircase',
 'staircase.core',
 'staircase.core.collections',
 'staircase.core.ops',
 'staircase.core.poly',
 'staircase.core.stats',
 'staircase.core.tools',
 'staircase.docstrings',
 'staircase.util']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=2',
 'numpy>=1.12,<2.0',
 'pandas>=0.24,<2',
 'pytz',
 'sortedcontainers>=2,<3']

setup_kwargs = {
    'name': 'staircase',
    'version': '1.6.3',
    'description': 'Facilitating the modelling, manipulation and analysis of data with (mathematical) step functions',
    'long_description': '<p align="center"><a href="https://github.com/venaturum/staircase"><img src="https://github.com/venaturum/staircase/blob/master/docs/img/staircase.png?raw=true" title="staircase logo" alt="staircase logo"></a></p>\n\n\n<p align="center">\n\t<a href="https://pepy.tech/project/staircase/" alt="PyPI downloads">\n        <img src="https://pepy.tech/badge/staircase" /></a>\n    <a href="https://www.python.org/" alt="Python version">\n        <img src="https://img.shields.io/pypi/pyversions/staircase" /></a>\n    <a href="https://pypi.org/project/staircase/" alt="PyPI version">\n        <img src="https://img.shields.io/pypi/v/staircase" /></a>\n    <a href="https://anaconda.org/conda-forge/staircase" alt="Conda Forge version">\n        <img src="https://anaconda.org/conda-forge/staircase/badges/version.svg?branch=master&kill_cache=1" /></a>\n    <a href="https://staircase.mit-license.org/" alt="License">\n        <img src="http://img.shields.io/:license-mit-blue.svg?style=flat-square"></a>\n</p>\n<p align="center">\n\t<a href="https://travis-ci.com/github/venaturum/staircase" alt"Travis CI">\n\t\t<img src="https://travis-ci.com/venaturum/staircase.svg"/></a>\n    <a href="https://railing.readthedocs.io/en/latest/" alt="Read the Docs">\n        <img src="https://readthedocs.org/projects/railing/badge/?version=latest" /></a>\n\t<a href="https://www.codacy.com/gh/venaturum/staircase/dashboard" alt="Codacy Grade">\n        <img src="https://app.codacy.com/project/badge/Grade/04eb9240feec4b4e95b20b96d060d38c" /></a>\t\n\t<a href="https://codecov.io/gh/venaturum/staircase"  alt="Codecov coverage">\n\t\t<img src="https://codecov.io/gh/venaturum/staircase/branch/master/graph/badge.svg"/></a>\n</p>\n<p align="center">\n\t<a href="https://mybinder.org/v2/gh/venaturum/staircase/master?filepath=docs%2Fexamples" alt="Binder">\n        <img src="https://mybinder.org/badge_logo.svg" /></a>\t\n\t<a href="https://colab.research.google.com/github/venaturum/staircase/blob/master/docs/examples/Index.ipynb" alt="Colab">\n        <img src="https://colab.research.google.com/assets/colab-badge.svg" /></a>\t\t\n</p>\n\nThe leading use-case for the staircase package is for the creation and analysis of step functions.\n\nPretty exciting huh.\n\nBut don\'t hit the close button on the browser just yet.  Let us convince you that much of the world around you can be modelled as step functions.\n\nFor example, the number of users viewing this page over time can be modelled as a step function.  The value of the function increases by 1 every time a user arrives at the page, and decreases by 1 every time a user leaves the page.  Let\'s say we have this data in vector format (i.e. tuple, list, numpy array, pandas series).  Specifically, assume *arrive* and *leave* are vectors of times, expressed as minutes past midnight, for all page views occuring yesterday.  Creating the corresponding step function is simple.  To achieve it we use the *[Stairs](https://railing.readthedocs.io/en/latest/Stairs.html)* class:\n\n```python\n>>> import staircase as sc\n\n>>> views = sc.Stairs()\n>>> views.layer(arrive,leave)\n```\n\nWe can visualise the function with the plot function:\n```python\n>>> views.plot()\n```\n<p align="left"><img src="https://github.com/venaturum/staircase/blob/master/docs/img/pageviews.png?raw=true" title="pageviews example" alt="pageviews example"></p>\n\nWe can find the total time in minutes the page was viewed:\n```python\n>>> views.integrate(0,1440)\n9297.94622521079\n```\n\nWe can find the average number of viewers:\n```python\n>>> views.mean(0,1440)\n6.4569071008408265\n```\n\nWe can find the average number of viewers, per hour of the day, and plot:\n```python\n>>> pd.Series([views.mean(60*i, 60*(i+1)) for i in range(24)]).plot()\n```\n<p align="left"><img src="https://github.com/venaturum/staircase/blob/master/docs/img/meanperhour.png?raw=true" title="mean page views per hour" alt="mean page views per hour"></p>\n\nWe can find the maximum concurrent views:\n```python\n>>> views.max(0,1440)\n16\n```\n\nWe can create histogram data showing relative frequency of concurrent viewers (and plot it):\n```python\n>>> views.hist(0,1440).plot.bar()\n```\n<p align="left"><img src="https://github.com/venaturum/staircase/blob/master/docs/img/pageviewshist.png?raw=true" title="concurrent viewers histogram" alt="concurrent viewers histogram"></p>\n\n\nPlotting is based on [matplotlib](https://matplotlib.org) and it requires relatively little effort to take the previous chart and improve the aesthetics:\n<p align="left"><img src="https://github.com/venaturum/staircase/blob/master/docs/img/pageviewshistpretty.png?raw=true" title="concurrent viewers histogram (aesthetic)" alt="concurrent viewers histogram (aesthetic)"></p>\n\n\nThere is plenty more analysis that could be done.  The staircase package provides a rich variety of [arithmetic operations](https://railing.readthedocs.io/en/latest/Stairs.html#arithmetic-operators), [relational operations](https://railing.readthedocs.io/en/latest/Stairs.html#relational-operators), [logical operations](https://railing.readthedocs.io/en/latest/Stairs.html#logical-operators), [statistical operations](https://railing.readthedocs.io/en/latest/Stairs.html#statistical-operators), for use with *Stairs*, in addition to functions for [univariate analysis](https://railing.readthedocs.io/en/latest/Stairs.html#summary-statistics), [aggregations](https://railing.readthedocs.io/en/latest/module_funcs.html) and compatibility with [pandas.Timestamp](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Timestamp.html).\n\n\n## Installation\n\nStaircase can be installed from PyPI:\n\n```bash\npython -m pip install staircase\n```\n\nor also with conda:\n\n```bash\nconda install -c conda-forge staircase\n```\n\n## Documentation\nThe complete guide to using staircase can be found at [Read the Docs](https://railing.readthedocs.io/en/latest/index.html)\n\n\n## Need help?\nPost your question on [Stack Overflow](https://stackoverflow.com/) and use the tag **staircase**.\n\n\n## Contributing\nThere are many ways in which contributions can be made - the first and foremost being *using staircase and giving feedback*.\n\nBug reports, feature requests and ideas can be submitted via the [Github issue tracker](https://github.com/venaturum/staircase/issues).\n\nAdditionally, bug fixes. enhancements, and improvements to the code and documentation are also appreciated and can be done via pull requests.\nTake a look at the current issues and if there is one you would like to work on please leave a comment to that effect.\n\nSee this [beginner\'s guide to contributing](https://github.com/firstcontributions/first-contributions), or [Pandas\' guide to contributing](https://pandas.pydata.org/pandas-docs/stable/development/contributing.html), to learn more about the process.\n\n\n## Versioning\n\nWe use [SemVer](http://semver.org/) for versioning. For the versions available, see the [tags on this repository](https://github.com/venaturum/staircase/tags). \n\n\n## License\n\nThis project is licensed under the MIT License - see the [LICENSE](https://staircase.mit-license.org/) file for details\n\n## Acknowledgments\n\n* This project is heavily reliant on [sorted containers](http://www.grantjenks.com/docs/sortedcontainers/).  Grant Jenks has done a great job bringing this functionality to Python at lightning fast speeds.\n* staircase began development from within the Hunter Valley Coal Chain Coordinator.  Thanks for the support!',
    'author': 'Riley Clement',
    'author_email': 'venaturum@gmail.com',
    'maintainer': 'Riley Clement',
    'maintainer_email': 'venaturum@gmail.com',
    'url': 'https://github.com/venaturum/staircase',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



number_of_steps_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s1.plot()
            >>> s1.number_of_steps()
            5
"""

min_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s2.plot()
            >>> s2.min(0, 3)
            0
            >>> s2.min(0, 3, upper_how='right')
            -1
            >>> s2.min(0, 2)
            0.5
            >>> s2.min(0, 2, lower_how='left')
            0.0
"""

max_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s1.plot()
            >>> s1.max()
            1.0
            >>> s1.max(4, 5)
            -1
            >>> s1.max(4, 5, lower_how='left')
            1
            >>> s1.max(4, 5, upper_how='right')
            0

"""

shift_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> stair_list = [s2, s2.shift(1), s2.shift(-1)]
            >>> fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(17,5), sharey=True, sharex=True)
            >>> for ax, title, stair_instance in zip(axes, ("s2", "s2.shift(1)", "s2.shift(-1)"), stair_list):
            ...     stair_instance.plot(ax, label=title)
            ...     ax.set_title(title)

        Note that the definition of shift is designed to be consistent with pandas.Series.shift

        .. plot::
            :context: close-figs

            >>> pd.Series(s2(range(7)))
            0    0.5
            1    0.5
            2    0.0
            3   -1.0
            4   -1.0
            5   -1.0
            6    0.0
            dtype: float64

            >>> pd.Series(s2(range(7))).shift(1)
            0    NaN
            1    0.5
            2    0.5
            3    0.0
            4   -1.0
            5   -1.0
            6   -1.0
            dtype: float64

            >>> pd.Series(s2.shift(1)(range(7)))
            0    0.0
            1    0.5
            2    0.5
            3    0.0
            4   -1.0
            5   -1.0
            6   -1.0
            dtype: float64
"""

diff_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> stair_list = [s2, s2.shift(1), s2.diff(1)]
            >>> fig, axes = plt.subplots(nrows=1, ncols=3, figsize=(17,5), sharey=True, sharex=True)
            >>> for ax, title, stair_instance in zip(axes, ("s2", "s2.shift(1)", "s2.diff(1)"), stair_list):
            ...     stair_instance.plot(ax, label=title)
            ...     ax.set_title(title)
            ... s2.plot(axes[1], label="s2", linestyle="--")
            ... axes[1].legend()

        Note that the definition of diff is designed to be consistent with pandas.Series.diff

        .. plot::
            :context: close-figs

            >>> pd.Series(s2(range(7)))
            0    0.5
            1    0.5
            2    0.0
            3   -1.0
            4   -1.0
            5   -1.0
            6    0.0
            dtype: float64

            >>> pd.Series(s2(range(7))).diff(1)
            0    NaN
            1    0.0
            2   -0.5
            3   -1.0
            4    0.0
            5    0.0
            6    1.0
            dtype: float64

            >>> pd.Series(s2.diff(1)(range(7)))
            0    0.5
            1    0.0
            2   -0.5
            3   -1.0
            4    0.0
            5    0.0
            6    1.0
            dtype: float64
"""

describe_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s2.plot()

        .. plot::
            :context: close-figs

            >>> s2.describe()
            unique    3.000000
            mean     -0.272727
            std       0.686349
            min      -1.000000
            25%      -1.000000
            50%       0.000000
            75%       0.500000
            max       0.500000
            dtype: float64

            >>> s2.describe(lower=0, upper=6, percentiles=range(0,101,20))
            unique    3.000000
            mean     -0.250000
            std       0.661438
            min      -1.000000
            0%       -1.000000
            20%      -1.000000
            40%      -1.000000
            60%       0.000000
            80%       0.500000
            100%      0.500000
            max       0.500000
            dtype: float64
"""

cov_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5), sharey=True, sharex=True)
            >>> for ax, title, stair_instance in zip(axes, ("s1", "s2"), (s1, s2)):
            ...     stair_instance.plot(ax, label=title)
            ...     ax.set_title(title)

        .. plot::
            :context: close-figs

            >>> s1.cov(s2)
            0.1404958677685951

            >>> s2.cov(s1)
            0.1404958677685951

            >>> s1.cov(s2, lower=0, upper=6)
            0.125

            >>> # autocovariance with lag 1
            >>> s1.cov(s1, lower=1, upper=5, lag=1)
            -0.3333333333333333

            >>> # cross-covariance with lag 1
            >>> s1.cov(s2, lower=1, upper=4.5, lag=1)
            0.15999999999999998

            >>> # cross-covariance with lag 1
            >>> s1.cov(s2, lower=1, upper=4.5, lag=1, clip='post')
            0.163265306122449

"""

corr_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5), sharey=True, sharex=True)
            >>> for ax, title, stair_instance in zip(axes, ("s1", "s2"), (s1, s2)):
            ...     stair_instance.plot(ax, label=title)
            ...     ax.set_title(title)

        .. plot::
            :context: close-figs

            >>> s1.corr(s2)
            0.24687803791136045

            >>> s2.corr(s1)
            0.24687803791136045

            >>> s1.corr(s2, lower=0, upper=6)
            0.27500954910846337

            >>> # autocorrelation with lag 1
            >>> s1.corr(s1, lower=1, upper=5, lag=1)
            -0.8660254037844386

            >>> # cross-correlation with lag 1
            >>> s1.corr(s2, lower=1, upper=5.5, lag=1)
            0.4961389383568339

            >>> # cross-correlation with lag 1
            >>> s1.corr(s2, lower=1, upper=4.5, lag=1, clip='post')
            0.4961389383568339
"""

rolling_mean_example = """
        Examples
        --------

        .. plot::
            :context: close-figs

            >>> s2.rolling_mean(window=[-0.5, 0.5])
            -0.5    0.0
             0.5    0.5
             1.5    0.5
             2.5    0.0
             3.5   -1.0
             5.0   -1.0
             6.0    0.0
            dtype: float64

        .. plot::
            :context: close-figs

            >>> series_list = [s2.rolling_mean(window=[-0.5, 0.5]), s2.rolling_mean(window=[-0.5, 0.5], lower=0, upper = 5.5)]
            >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5), sharey=True, sharex=True)
            >>> for ax, title, series in zip(axes, ("default", "bounds supplied"), series_list):
            ...     s2.plot(ax=ax)
            ...     series.plot(ax=ax, label='rolling mean')
            ...     ax.set_title(title)
            ...     ax.legend()

        .. plot::
            :context: close-figs

            >>> series_list = [s2.rolling_mean(window=[-1, 0], lower=0, upper = 5.5), s2.rolling_mean(window=[0, 1], lower=0, upper = 5.5)]
            >>> fig, axes = plt.subplots(nrows=1, ncols=2, figsize=(12,5), sharey=True, sharex=True)
            >>> for ax, title, series in zip(axes, ("trailing window", "leading window"), series_list):
            ...     s2.plot(ax=ax)
            ...     series.plot(ax=ax, label='rolling mean')
            ...     ax.set_title(title)
            ...     ax.legend()
"""
